from .TrueCompression import GearLlamaForCausalLMNew
from .TrueCompression import GearLlamaForCausalLM

from .Simulated import CompressionConfig
from .Simulated import SimulatedGearLlamaForCausalLM

# from .modeling_llama_h2o import H2OLlamaForCausalLM, LlamaConfig
from .Simulated import SimulatedGearMistralForCausalLM, MistralConfig
from .Simulated import LlamaForCausalLMH2O
