from .compress_config import CompressionConfig
from .modeling_llama_new import SimulatedGearLlamaForCausalLM

# from .modeling_llama_h2o import H2OLlamaForCausalLM, LlamaConfig
from .modeling_mistral import SimulatedGearMistralForCausalLM, MistralConfig
from .h2o_llama_self_written import LlamaForCausalLMH2O
